(** The classic atoms found in most logics (propositional logic, predicate logic, etc.). *)

module type PARAM =
  sig
    val names : string list (** The list of admissible names for atoms. If this list is empty, every name is admissible (open set of atoms). *)
  end

module Make (Param : PARAM) : Logic.T =
  struct
    include Logic.Default

    type t = Attr of string | Term of string

    let isvalue _ = true

    let parse = parser
      |	[<'Token.Term n
	  when (Param.names = [] or List.mem n Param.names)
	      & not (List.mem n Syntax.keywords)>] -> Term n
      |	[<'Token.Ident a
	  when (Param.names = [] or List.mem a Param.names)
	      & not (List.mem a Syntax.keywords)>] -> Attr a

    let parse_compact = parser
      | [<'Token.String s>] ->
          let res = ref [] in
          let l = String.length s in
          for i = l - 1 downto 0 do
            match s.[i] with
            | 'a' .. 'z' -> res := Token.Ident (String.sub s i 1) :: !res
            | 'A' .. 'Z' -> res := Token.Term (String.sub s i 1) :: !res
            | ' ' | '\t' | '\n' -> res := Token.Term "_" :: !res
            | _ -> ()
          done;
          !res

    let print = function
      | Attr a -> [Token.Ident a]
      |	Term name -> [Token.Term name]

    let rec entails a b = a=b

    let conj a b =
      if entails a b then a 
      else if entails b a then b
      else raise Not_found

    let add a b = b

    let sub a b = if entails a b then raise Not_found else a

    let features a = []

    let rec gen y d gs =
      if y=d & not (List.exists (fun x -> x = y) gs)
      then LSet.singleton y
      else LSet.empty ()
  end

