(** Interpretations (semantics) are sets of interpretations from the sub-logic. *)

module type PARAM =
  sig
    val toks : Syntax.t_list (** List of tokens to be used as a prefix of formulas. *)
  end

module Make (Param : PARAM) (E : Logic.T) : Logic.T =
  struct
    include Logic.Default

    type t = E.t

    let parse str =
      if Param.toks = []
      then match str with parser
        [<e = E.parse>] -> e
      else match str with parser
        [<_ = Syntax.parse_tokens Param.toks;
          e = E.parse ?? "Syntax error after: " ^ Syntax.stringizer Param.toks
         >] -> e

    let print e = Param.toks @ E.print e

    let simpl e = E.simpl e

    let top () = E.top ()

    let bot () = E.bot ()

    let entails f g = E.entails f g

    let conj f g =
      if entails f g then f
      else if entails g f then g
      else raise Not_found

    let add f g = E.add f g

    let sub f g = E.sub f g

    let features f = E.features f

    let axiom f g = E.axiom f g

    let gen f g hs = E.gen f g hs

  end
