(** Program for generating sum logics of arbitrary arity. *)

open Printf

let _ =
  try
    let n = int_of_string (Sys.argv.(1)) in
    let ch = open_out ("sum" ^ string_of_int n ^ ".ml") in

    fprintf ch "module Make";
    for i = 1 to n do fprintf ch " (A%u : Logic.T)" i done;

    fprintf ch " (Trad :\n";
    fprintf ch "sig\n";
    fprintf ch "\ttype t = Top ";
    for i = 1 to n do fprintf ch "| A%u of A%u.t " i i done;
    fprintf ch "\n\tval trad : t * t -> t * t\n";
    fprintf ch "end) : Logic.T =\n";

    fprintf ch "struct\n";
    fprintf ch "include Logic.Default\n";
    fprintf ch "type t = Trad.t\n\n";
    fprintf ch "open Trad\n\n";

    fprintf ch "let isvalue = function\n";
    fprintf ch "| Top -> false\n";
    for i = 1 to n do fprintf ch "| A%u x -> A%u.isvalue x\n" i i done;
    fprintf ch "\n";

    fprintf ch "let top () = Top\n";

    fprintf ch "let bot () =\n";
    for i = 1 to n-1 do fprintf ch "try A%u (A%u.bot ()) with Not_found ->\n" i i done;
    fprintf ch "A%u (A%u.bot ())\n\n" n n;

    fprintf ch "let entails f g = match trad (f,g) with\n";
    fprintf ch "| _, Top -> true\n";
    for i = 1 to n do fprintf ch "| A%u x, A%u y -> A%u.entails x y\n" i i i done;
    fprintf ch "| _, _ -> false\n\n";

    fprintf ch "let conj f g = match trad (f,g) with\n";
    fprintf ch "| Top, _ -> g\n";
    fprintf ch "| _, Top -> f\n";
    for i = 1 to n do fprintf ch "| A%u x, A%u y -> A%u (A%u.conj x y)\n" i i i i done;
    fprintf ch "| _, _ -> raise Not_found\n\n";

    fprintf ch "let add f g = conj f g\n\n";

    fprintf ch "let sub f g = if entails f g then raise Not_found else f\n\n";

    fprintf ch "let features = function\n";
    fprintf ch "| Top -> []\n";
    for i = 1 to n do fprintf ch "| A%u x -> List.map (fun (vis,x) -> vis,A%u x) (A%u.features x)\n" i i i done;
    fprintf ch "\n";

    fprintf ch "let axiom f g_opt = match g_opt with\n";
    fprintf ch "| None\n";
    fprintf ch "| Some Top -> (match f with\n";
    for i = 1 to n do
      fprintf ch "| A%u x -> A%u.axiom x None\n" i i;
    done;
    fprintf ch "| _ -> LSet.empty ())\n";
    fprintf ch "| Some g -> (match trad (f,g) with\n";
    for i = 1 to n do
      fprintf ch "| A%u x, A%u y -> A%u.axiom x (Some y)\n" i i i
    done;
    fprintf ch "| _, _ -> LSet.empty ())\n\n";

    fprintf ch "let gen f g hs = match trad (f,g) with\n";
    for i = 1 to n do fprintf ch "| A%u x, A%u y -> List.map (fun t -> A%u t) (A%u.gen x y (List.fold_right (fun h zs -> match trad (A%u x,h) with A%u x, A%u z -> z::zs | _, _ -> zs) hs []))\n" i i i i i i i done;
    fprintf ch "| _, _ -> []\n\n";

    fprintf ch "let parse = parser\n";
    fprintf ch "| [<'Token.Interro>] -> Top\n";
    for i = 1 to n do fprintf ch "| [<x = A%u.parse>] -> A%u x\n" i i done;
    fprintf ch "\n";

    fprintf ch "let print = function\n";
    fprintf ch "| Top -> [Token.Interro]\n";
    for i = 1 to n do fprintf ch "| A%u x -> A%u.print x\n" i i done;
    fprintf ch "\n";

    fprintf ch "let simpl = function\n";
    fprintf ch "| Top -> [<>]\n";
    for i = 1 to n do fprintf ch "| A%u x -> Common.stream_map (fun x -> A%u x) (A%u.simpl x)\n" i i i done;
    fprintf ch "\n";

    fprintf ch "end\n"
  with
    Invalid_argument _ -> print_endline "Use: make_sum <int>"
  | Sys_error msg -> print_endline ("Error: " ^ msg)
