(** Make optional the use of a sub-logic formula. *)

module type PARAM =
  sig
    val none : Syntax.t_list (** List of tokens used to represent the 'void' case. *)
  end

module Make (Param : PARAM) (A : Logic.T) : Logic.T =
  struct
    include Logic.Default

    type t = A.t option

    let isvalue = function
      | None -> false
      | Some x -> A.isvalue x

    let top () = None

    let bot () = Some (A.bot ())

    let entails f g = match f, g with
      _, None -> true
    | None, Some _ -> false
    | Some x, Some y -> A.entails x y

    let conj f g = match f, g with
      f, None -> f
    | None, g -> g
    | Some x, Some y -> Some (A.conj x y)

    let add f g = conj f g

    let sub f g = match (f,g) with
      None, Some _ -> None
    | Some x, Some y -> Some (A.sub x y)
    | _, _ -> raise Not_found

    let features = function
	None -> [true,None]
      |	Some x -> (true,None)::List.map (fun (vis,x) -> vis,Some x) (A.features x)

    let axiom f g_opt =
      match f, g_opt with
      | Some x, None -> A.axiom x None
      | Some x, Some (None) -> A.axiom x None
      | Some x, Some (Some y) -> A.axiom x (Some y)
      | _, _ -> LSet.empty ()

    let gen f g hs =
    if false & hs = [] then [None]
    else match (f,g) with
    | Some x, Some y ->
         let zs = Common.mapfilter (function x -> x) hs in
         List.fold_left (fun res x -> LSet.add (Some x) res) (LSet.empty ()) (A.gen x y zs)
    | _, _ -> [None]

    let parse = parser
      | [<x = A.parse>] -> Some x
      |	[<_ = Syntax.parse_tokens Param.none>] -> None

    let print = function
	None -> Param.none
      |	Some x -> A.print x

    let simpl = function
      | None -> [<>]
      | Some x -> [<'None; Common.stream_map (fun a -> Some a) (A.simpl x)>]

  end
